# This file is part of PyCm, a Python interface to the Cm library
#
# Copyright (C) 2020-2021  European Gravitational Observatory
#
# Author list: Franco Carbognani: franco.carbognani@ego-gw.it
#                   Rhys Poulton: poulton@ego-gw.it
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import os
import sys
import subprocess

headers = {}
headers["cm"] = ["CmMessage.h","CmConnect.h","CmTransaction.h"]
headers["pycm"] = ["Handlers.h"]
pycm_library = "libpycm.so"
output_name = "PyCm/PyCm.py"

#Check if this script is being used directly by the user
if sys.stdin.isatty():
   print("It is suggested that the user not run this scripts as it will be run when this project is built via cmake (which provides the correct location for libraries).")

if(len(sys.argv)<3):
   raise SystemExit("""ERROR: Incorrect number of arguments passed to the build_ctypesgen.py script
This script requires arguments: /path/to/c/compiler /path/to/PyCm/<builddir>/src/, /path/to/libPyCm.so/install/dir/, /path/to/PyCm/src/ and  /path/to/cm/include/
""")


c_compiler = sys.argv[1]
pycm_library_dir = sys.argv[2]
install_library_dir = sys.argv[3]
include_dirs = sys.argv[4:]

def check_executable(executable):
   """ Check if executable can be executed
   """
   is_executable = os.access(executable, os.X_OK)
   is_not_directory = os.path.isfile(executable)
   if is_executable and is_not_directory:
       return True
   else:
       return False


def cmd_exists(cmd, path=None):
    """ test if path contains an executable file with name 
    """
    if path is None:
        path = os.environ["PATH"].split(os.pathsep)
    for prefix in path:
        filename = os.path.join(prefix, cmd)
        if(check_executable(filename)):
            return True
    return False

#Try and find the ctypesgen
ctypegen_exe = os.getenv("CTYPESGENEXE")
if(ctypegen_exe):
    if(check_executable(ctypegen_exe)==False):
       raise SystemExit("The CTYPESGENEXE environmental varible is set but is not a executable")
    else:
       print("Using user defined ctypesgen executable:",ctypegen_exe)
else:
    #Check if there is a local installation of ctypesgen in .local/bin
    local_path = os.path.join(os.path.expanduser("~"),".local/bin")
    ctypegen_exe = "ctypesgen"
    if(cmd_exists(ctypegen_exe,[local_path])):
        ctypegen_exe = os.path.join(local_path,ctypegen_exe)
        print("Found local installation of ctypegen at:",ctypegen_exe)
    elif(cmd_exists(ctypegen_exe)):
        print("Found system installation of ctypesgen")
    else:
        print("""
WARNING: Installation of ctypesgen not found please either
   - Install via 'pip install ctypesgen (--user)'
   - Set CTYPESGENEXE environmental varible to point to the ctypesgen executable
Installation will continue using pre-generated PyCm script in PyCm/PyCm.py, which can potentially can have incompatible verison of PyCm""")
        raise SystemExit()

#Try and find the header files for the libraries included
include_dir_dict = {"cm":"","pycm":""}
all_headers = []
for libname in include_dir_dict:

    for include_dir in include_dirs:
        filename = os.path.join(include_dir,headers[libname][0])
        if os.path.isfile(filename):
            include_dir_dict[libname] = include_dir
            all_headers.extend([os.path.join(include_dir,header) for header in headers[libname]])
            break
    #If normally exited from for loop then there is a library/ header file missing
    else:
        raise SystemExit("Error could not find the header file %s for library %s" %(headers[libname][0],libname))

#Generate
cmd = ctypegen_exe + " --cpp '" + c_compiler+ " -E' " +\
      " -a -l " + pycm_library +\
      " --compile-libdir " + pycm_library_dir +\
      " --runtime-libdir "+ install_library_dir +\
      " -I " + " -I ".join(include_dir_dict.values()) +\
      " " + " ".join(all_headers) +\
      " -o " + output_name
output = subprocess.check_output(cmd,shell=True)

#Get the license info
licence_header_file = open("scripts/license_header.txt","r")
licence_header = licence_header_file.readlines()
licence_header_file.close()

#Add the addtional header functions for c_void and POINTER
addtional_preamble_file = open("scripts/additional_preamble.txt","r")
addtional_preamble = addtional_preamble_file.read()
addtional_preamble_file.close()

#Now edit the generated file to fix the known issues
with open(output_name,"r+") as output_file:

   content =  output_file.readlines()
   for i in range(len(content)):
       if("from ctypes import *" in content[i]):
           content[i] = "from ctypes import *" + addtional_preamble
       elif("argtypes" in content[i] and "CmMessageHandler" in content[i]):
           content[i] = content[i].replace("CmMessageHandler","POINTER(CmMessageHandler)")
       elif("raise" in content[i] and "=" in content[i]):
           content[i] = content[i].replace("raise","_raise",1)


   #Re-write the binding with corrections
   output_file.seek(0)
   output_file.writelines(licence_header + content)


